#!/usr/bin/env python3
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""Make a standalone .rules file."""
from xrhardware.db import get_devices
from xrhardware.generate import generate_rules_file_contents

if __name__ == "__main__":
    all_rules = '\n\n'.join(d.make_commented_rule() for d in get_devices())
    contents = '\n'.join((
        '# BEGIN DEVICE LIST #',
        '#####################',
        all_rules,
        '###################',
        '# END DEVICE LIST #',
    ))
    print(generate_rules_file_contents(
        __file__, "70-xrhardware.rules", contents))



#!/usr/bin/env python3 -i
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""XR Hardware device database element type."""

import attr
import re

_VID_PID_RE = re.compile(r'^[0-9a-f]{4}$')

_USBSERIALNAME_RE = re.compile(r'^[._0-9a-zA-Z]+$')


def _is_four_hexits(_self, _attribute, val):
    if not _VID_PID_RE.match(val):
        raise ValueError("Must be four hexadecimal characters in lowercase")


def _is_suitable_device_name(_self, _attribute, val):
    if val is None:
        return
    if not _USBSERIALNAME_RE.match(val):
        raise ValueError(
            "Must be a reasonable thing to put in a device node name: "
            "a-z, A-Z, numbers, ., _")


@attr.s
class Device:
    """An XR hardware device that we should permit access to."""

    description = attr.ib()
    vid = attr.ib(default=None, validator=_is_four_hexits)
    pid = attr.ib(default=None, validator=_is_four_hexits)
    usb_serial_name = attr.ib(default=None, validator=_is_suitable_device_name)
    usb = attr.ib(default=True)
    bluetooth = attr.ib(default=False)
    extra_properties = attr.ib(default=None)

    def get_properties_to_set(self):
        """Generate the udev properties to set for this device."""
        yield ('ID_xrhardware', 1)

        if self.usb_serial_name:
            yield ('ID_xrhardware_USBSERIAL_NAME', self.usb_serial_name)
        if self.extra_properties:
            for k, v in self.extra_properties.items():
                yield (k, v)

    def yield_hwdb_identification(self):
        """Compute identification for hwdb recognition of this device."""
        if not self.vid:
            raise RuntimeError(
                "Can't make a hwdb entry for something without a vid!")
        # usb: and bluetooth: prefix come later
        parts = []
        parts.append('v%s' % self.vid)
        if self.pid:
            parts.append('p%s' % self.pid)
        parts.append('*')
        identifier_suffix = ''.join(parts)
        if self.usb:
            yield 'usb:' + identifier_suffix
        if self.bluetooth:
            yield 'bluetooth:' + identifier_suffix

    def make_hwdb_entry(self):
        """Return a hwdb entry for this device."""
        lines = ['# ' + self.description]
        lines.extend(list(self.yield_hwdb_identification()))
        for k, v in self.get_properties_to_set():
            lines.append(' %s=%s' % (k, v))
        return '\n'.join(lines)

    @property
    def extended_description(self):
        """Return the description augmented with the interface types."""
        interfaces = []
        if self.bluetooth:
            interfaces.append("Bluetooth")
        if self.usb:
            interfaces.append("USB")
        return "%s - %s" % (self.description, ", ".join(interfaces))

    def make_commented_rule(self):
        """Return a comment and udev rule for this device."""
        return '\n'.join((
            "# %s" % self.extended_description,
            self.make_rule(),
            ""))

    def yield_rule_condition_lists(self):
        """Yield the udev rule conditions (as lists) to select this device."""
        if self.usb:
            parts = []
            if self.vid:
                parts.append('ATTRS{idVendor}=="%s"' % self.vid)
            if self.pid:
                parts.append('ATTRS{idProduct}=="%s"' % self.pid)
            yield parts

        if self.bluetooth:
            yield ['KERNELS=="0005:%s:%s.*"' %
                   (self.vid.upper(), self.pid.upper())]
            # Bluetooth devices don't get idVendor and idProduct,
            # but they do get this which is pretty similar.
            # However, these fail to trigger the rule properly for the PS Move
            # at least.
            #
            # parts = ['ATTRS{id/bustype}=="0005"']
            # if self.vid:
            #     parts.append('ATTRS{id/vendor}=="%s"' % self.vid)
            # if self.pid:
            #     parts.append('ATTRS{id/product}=="%s"' % self.pid)
            # yield parts

    def yield_rule_conditions(self):
        """Yield the udev rule conditions to select this device."""
        for conditions in self.yield_rule_condition_lists():
            yield ", ".join(conditions)

    def make_rule(self):
        """Return a udev rule for this device."""
        rules = []
        for condition in self.yield_rule_conditions():
            parts = [condition, 'TAG+="uaccess"']
            parts.extend('ENV{%s}="%s"' % (k, v)
                         for k, v in self.get_properties_to_set())
            rules.append(', '.join(parts))
        return "\n".join(rules)



#!/usr/bin/env python3
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""XR Hardware generation helper code."""

# pylint: disable=E501
_SHARED_PREFIX = """# Do not edit this file - generated by {script}
# SPDX-License-Identifier: BSL-1.0
"""


_RULES_TEMPLATE_PREFIX = """
# Must be located at a number smaller than 73 for uaccess to work right!

# Skip if a remove
ACTION=="remove", GOTO="xrhardware_end"

"""

# Must be before 73 to have uaccess do what it should!
_RULES_TEMPLATE_SUFFIX = """

# Exit if we didn't find one
ENV{ID_xrhardware}!="1", GOTO="xrhardware_end"

# XR devices with serial ports aren't modems, modem-manager
ENV{ID_xrhardware_USBSERIAL_NAME}!="", SUBSYSTEM=="usb", ENV{ID_MM_DEVICE_IGNORE}="1"

# Make friendly symlinks for XR USB-Serial devices.
ENV{ID_xrhardware_USBSERIAL_NAME}!="", SUBSYSTEM=="tty", SYMLINK+="ttyUSB.$env{ID_xrhardware_USBSERIAL_NAME}"

LABEL="xrhardware_end"
"""  # noqa: E501 # it's too long because that's what the template is.


def generate_file_contents(script, body):
    """Return generated file contents for the given script file and body."""
    from pathlib import Path
    script = str(Path(script).name)
    shared_prefix = _SHARED_PREFIX.format(script=script)
    return '\n'.join((shared_prefix, body))


def generate_rules_file_contents(script, fn, body):
    """Return the body surrounded by the rules template."""
    parts = ['# ' + fn, _RULES_TEMPLATE_PREFIX, body, _RULES_TEMPLATE_SUFFIX]
    return generate_file_contents(script, "\n".join(parts))



#!/usr/bin/env python3
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""XR Hardware device database consistency test."""
from xrhardware.db import get_devices

if __name__ == "__main__":
    print("Testing device database entries' consistency and uniqueness")
    known = set()
    import re
    hexstring = re.compile(r"[0-9a-fA-F]{4}")
    for dev in get_devices():

        print("Checking entry for:", dev.extended_description)

        # Duplicate ID?
        for dev_id in dev.yield_hwdb_identification():
            assert(dev_id not in known)
            known.add(dev_id)

        # Must be at least one of these
        assert(dev.usb or dev.bluetooth)

        # VID/PID must fit the expected format
        if dev.pid:
            assert(hexstring.match(dev.pid))
            # prefer lowercase for compatibility
            assert(dev.pid == dev.pid.lower())
        if dev.vid:
            assert(hexstring.match(dev.vid))
            # prefer lowercase for compatibility
            assert(dev.vid == dev.vid.lower())
    print("Success!")



#!/usr/bin/env python3
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""Make a hwdb file. Requires an associated .rules file."""

from xrhardware.db import get_devices
from xrhardware.generate import generate_file_contents

if __name__ == "__main__":
    all_entries = '\n\n'.join(d.make_hwdb_entry() for d in get_devices())
    contents = '\n'.join((
        "# This hwdb file must be used with the corresponding rules file.",
        all_entries
    ))
    print(generate_file_contents(__file__, contents))



#!/usr/bin/env python3
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""Make a .rules file for use with a hwdb file."""

from xrhardware.generate import generate_rules_file_contents

if __name__ == "__main__":

    print(generate_rules_file_contents(
        __file__,
        "70-xrhardware-hwdb.rules",
        '# This rules file must be used with the corresponding hwdb file'))



#!/usr/bin/env python3 -i
# Copyright 2019-2020 Collabora, Ltd
# SPDX-License-Identifier: BSL-1.0
# Author: Ryan Pavlik <ryan.pavlik@collabora.com>
"""XR Hardware device database."""

from .device import Device

# Please make sure VID/PID are lowercase.


def get_devices():
    """Return the entire device database in one really large tuple."""
    return (
        Device("Razer Hydra", "1532", "0300"),
        Device("HTC Vive", "0bb4", "2c87"),
        Device("HTC Vive Pro", "0bb4", "0309"),
        Device("Valve Watchman Dongle", "28de", "2101"),
        Device("Valve VR Radio", "28de", "2102"),
        Device("Valve Index Controller", "28de", "2300"),
        # Name might be wrong.
        Device("Valve Receiver for Lighthouse - HTC Vive", "28de", "2000"),
        # also PID 2220 (lighthouse fpga rx), 2300 (lhr vive pro?)
        Device("OSVR HDK", "1532", "0b00", usb_serial_name="OSVRHDK"),
        Device("OSVR HDK Camera", "0bda", "57e8"),
        Device("Sensics zSight", "16d0", "0515", usb_serial_name="zSight"),

        # Re-used ID from ST Micro "LED badge -- mini LED display -- 11x44"
        Device("NOLO CV1", "0483", "5750"),
        # Unknown where this VID was assigned
        Device("NOLO CV1", "28e9", "028a"),

        # ST Micro VID, and PID not listed in usb-ids so may be uniquely
        # assigned (or may be just copied from PID of DK2 since the product
        # string is "Rift P1")
        Device("Pimax 4k", "0483", "0021"),

        Device("Oculus Rift (DK1)", "2833", "0001"),
        Device("Oculus Rift (DK2)", "2833", "0021"),
        Device("Oculus Rift (DK2)", "2833", "2021"),
        Device("Oculus Rift (CV1)", "2833", "0031"),
        Device("Oculus Rift S", "2833", "0051"),
        Device("Samsung GearVR (Gen1)", "04e8", "a500"),

        Device("3Glasses-D3V1", "2b1c", "0200"),
        Device("3Glasses-D3V2", "2b1c", "0201"),
        Device("3Glasses-D3C", "2b1c", "0202"),
        Device("3Glasses-D2C", "2b1c", "0203"),
        Device("3Glasses-S1V5", "2b1c", "0100"),
        Device("3Glasses-S1V8", "2b1c", "0101"),

        # Need to be ignored by libinput to avoid spurious touchscreen events
        Device("Microsoft Windows MR Controller", "045e", "065b",
               bluetooth=True, usb=False,
               extra_properties={"LIBINPUT_IGNORE_DEVICE": "1"}),
        Device("Microsoft Windows MR Controller", "045e", "065d",
               bluetooth=True, usb=False,
               extra_properties={"LIBINPUT_IGNORE_DEVICE": "1"}),

        Device("Sony PlayStation VR", "054c", "09af"),

        Device("Sony PlayStation Move Motion Controller CECH-ZCM1",
               "054c", "03d5", bluetooth=True, usb=True),
        Device("Sony PlayStation Move Motion Controller CECH-ZCM2",
               "054c", "0c5e", bluetooth=True, usb=True),
        # TODO Duplicate of NOLO CV1?
        # Device("Deepoon", "0483", "5750"),

        Device("LG 360 VR R-100", "1004", "6374"),

        Device("Microsoft HoloLens Sensors", "045e", "0659"),
        Device("Samsung Odyssey+ sensors", "04e8", "7312"),
        Device("HP Reverb G1", "03f0", "0c6a"),
        Device("HP Reverb G2", "03f0", "0580"),

        # Pretends to be a DK1...
        # Device("VR-Tek WVR1", "2833", "0001"),

        # Vis3r NxtVR
        Device("Vis3r NxtVR", "1209", "9d0f"),
    )
